unit NConf;

{$IMPORTEDDATA OFF}

interface

uses NTypes, NLib, NDict;

type
  TConf = class(TStrArrayDict)
  public
    Parser: function (const S: string): TStringArray;
    class function ReadLine(const F: TStream; var Line: string): Boolean;
    function GetBool(const Key: string; Index: Integer=0;
        Default: Boolean=False): Boolean;
    function GetFloat(const Key: string; Index: Integer=0; Default: Double=0):
        Double;
    function GetInt(const Key: string; Index: Integer=0; Default: Integer=0):
        Integer;
    function GetStr(const Key: string; Index: Integer=0; Default: string=''):
        string;
    function GetStrW(const Key: string; Index: Integer=0; Default: WideString=''):
        WideString;
    procedure Read(const F: TStream); overload;
    procedure Read(const S: string); overload;
    procedure Write(const F: TStream);
    procedure LoadFromFile(const Path: WideString);
    procedure SaveToFile(const Path: WideString);
  end;

implementation

uses NSys;

class function TConf.ReadLine(const F: TStream; var Line: string): Boolean;
begin
  while F.ReadLine(Line) do begin
    RemoveLineSep(Line);
    if (line = '') or (Line[1] = '#') then
      Continue;
    Result := True;
    Exit;
  end;
  Result := False;
end;

function TConf.GetBool(const Key: string; Index: Integer; Default: Boolean):
    Boolean;
var
  args: TStringArray;
begin
  args := Items[Key];
  if High(args) < Index then
    Result := Default
  else begin
    try
      Result := (args[Index] = 'on') or (Sint(args[Index]) <> 0);
    except
      on EValueError do Result := False;
    end;
  end;
end;

function TConf.GetFloat(const Key: string; Index: Integer; Default: Double):
    Double;
var
  args: TStringArray;
begin
  args := Items[Key];
  if High(args) < Index then
    Result := Default
  else begin
    try
      Result := SFloat(args[Index]);
    except
      on EValueError do Result := 0;
    end;
  end;
end;

function TConf.GetInt(const Key: string; Index: Integer; Default: Integer):
    Integer;
var
  args: TStringArray;
begin
  args := Items[Key];
  if High(args) < Index then
    Result := Default
  else begin
    try
      Result := SInt(args[Index]);
    except
      on EValueError do Result := 0;
    end;
  end;
end;

function TConf.GetStr(const Key: string; Index: Integer; Default: string):
    string;
var
  args: TStringArray;
begin
  args := Items[Key];
  if High(args) < Index then
    Result := Default
  else
    Result := args[Index];
end;

function TConf.GetStrW(const Key: string; Index: Integer=0; Default: WideString=''):
    WideString;
var
  args: TStringArray;
begin
  args := Items[Key];
  if High(args) < Index then
    Result := Default
  else
    Result := DecodeUTF8(args[Index]);
end;

procedure TConf.LoadFromFile(const Path: WideString);
var
  f: TFile;
begin
  try
    f := TFile.Open(Path);
    try
      Read(f);
    finally
      f.Free;
    end;
  except
    on E: EFileNotFound do ;
  end;
end;

procedure TConf.SaveToFile(const Path: WideString);
var
  f: TFile;
begin
  f := TFile.Create(Path);
  try
    Write(f);
  finally
    f.Free;
  end;
end;

procedure TConf.Read(const F: TStream);
var
  line: string;
  args: TStringArray;
begin
  if not Assigned(Parser) then
    Parser := SplitCommandLine;
  args := nil;
  while TConf.ReadLine(F, line) do begin
    args := Parser(line);
    Items[args[0]] := SliceArray(args, 1);
  end;
end;

procedure TConf.Read(const S: string);
var
  bs: TBytesStream;
begin
  bs := TBytesStream.Create(S);
  try
    Read(bs);
  finally
    bs.Free;
  end;
end;

procedure TConf.Write(const F: TStream);
var
  i: Integer;
  s, key: string;
  args: TStringArray;
begin
  args := nil;
  while EachKeys(key) do begin
    args := Items[key];
    s := key;
    for i := 0 to High(args) do
      s := s + ' ' + args[i]; // TODO: support quote
    Print(F, s);
  end;
end;

end.
